/*
 * glFeedback, an OpenGL Feedback processing Library
 * Copyright (C) 2003 Francesco Bradascio
 *
 * E-mail: fbradasc@katamail.com
 * URL: http://digilander.libero.it/fbradasc/GL_FEEDBACK.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This work is mostly based on the GL2PS Library which is:
 *
 * Copyright (C) 1999-2003 Christophe Geuzaine
 * E-mail: geuz@geuz.org
 * URL: http://www.geuz.org/gl2ps/
 *
 * GL2PS Contributor(s):
 *   Michael Sweet <mike@easysw.com>
 *   Marc Ume <marc.ume@digitalgraphics.be>
 *   Jean-Francois Remacle <remacle@scorec.rpi.edu>
 *   Bart Kaptein <B.L.Kaptein@lumc.nl>
 *   Quy Nguyen-Dai<quy@vnilux.com>
 *   Sam Buss <sbuss@ucsd.edu>
 *   Shane Hill <Shane.Hill@dsto.defence.gov.au>
 *   Romain Boman <r_boman@yahoo.fr>
 *   Rouben Rostamian <rostamian@umbc.edu>
 *   Diego Santa Cruz <Diego.SantaCruz@epfl.ch>
 *   Shahzad Muzaffar <Shahzad.Muzaffar@cern.ch>
 *   Lassi Tuura <lassi.tuura@cern.ch>
 *   Guy Barrand <barrand@lal.in2p3.fr>
 */

#include <GLF/GLFBspTree.h>

GLFBspTree::GLFBspTree (
    GLint maxBestRoot,
    GLFPrimitives &prims,
    GLFPrimitives *storage
)
: primitives(storage) // the primitives list contains pointers to real
                      // primitives which are stored into the storage list
{
GLFPrimitive *frontprim, *backprim;
GLFPrimitives frontlist(storage), backlist(storage);
GLint numprimitives = prims.size();

    GLFPrimitive &prim = prims.findRoot(maxBestRoot);
    front              = NULL;
    back               = NULL;
    plane              = prim.plane();
    this->storage      = storage;

    primitives.add(prim);
    for (int i=0; i<numprimitives; i++) {
        GLFPrimitive &prim1=prims[i];
        if (&prim1!=&prim) {
            switch (prim1.classify(plane)) {
                case GLF_COINCIDENT:
                    primitives.add(prim1);
                break;
                case GLF_IN_BACK_OF:
                    backlist.add(prim1);
                break;
                case GLF_IN_FRONT_OF:
                    frontlist.add(prim1);
                break;
                case GLF_SPANNING:
                    backprim  = prim1.backPiece(plane);
                    frontprim = prim1.frontPiece(plane);
                    backlist.add(backprim);
                    frontlist.add(frontprim);
                    delete backprim;
                    delete frontprim;
                break;
            }
        }
    }
    if (!primitives.empty()) {
        primitives.sort(GLFPrimitive::ltcType);
    }
    if (!frontlist.empty()) {
        // frontlist.sort(GLFPrimitive::ltcType);
        front = new GLFBspTree(maxBestRoot, frontlist, this->storage);
    }
    if (!backlist.empty()) {
        // backlist.sort(GLFPrimitive::ltcType);
        back = new GLFBspTree(maxBestRoot, backlist, this->storage);
    }
}

GLFBspTree::~GLFBspTree()
{
    if (back) {
        delete back;
    }
    if (front) {
        delete front;
    }
    primitives.clear();
}

void GLFBspTree::traverse (
    const GLFxyz &eye,
    GLfloat epsilon,
    GLboolean (*compare)(GLfloat f1, GLfloat f2),
    void (*action)(GLFPrimitive& prim, void *data),
    void *data
) {
    GLfloat result = plane.compare(eye);

    if (compare(result, epsilon)) {
        if (back) {
            back->traverse(eye, epsilon, compare, action, data);
        }
        primitives.traverse(action, data);
        if (front) {
            front->traverse(eye, epsilon, compare, action, data);
        }
    } else if (compare(-epsilon, result)) {
        if (front) {
            front->traverse(eye, epsilon, compare, action, data);
        }
        primitives.traverse(action, data);
        if (back) {
            back->traverse(eye, epsilon, compare, action, data);
        }
    } else {
        if (front) {
            front->traverse(eye, epsilon, compare, action, data);
        }
        if (back) {
            back->traverse(eye, epsilon, compare, action, data);
        }
    }
}

#if 0
void GLFBspTree::addBoundaryInList(GLFPrimitive& prim)
{
GLshort i,n;
GLFxyz c;

    c[0] = c[1] = c[2] = 0.;
    for (i = 0; i < prim.verts.size(); i++) {
        c[0] += prim.verts[i].xyz[0];
        c[1] += prim.verts[i].xyz[1];
    }
    c[0] /= prim.verts.size();
    c[1] /= prim.verts.size();

    for (i = 0; i < prim.verts.size(); i++) {
        if (prim.attributes.polygon.boundary & (1 << i)) {
            GLFLine& b = (GLFLine&)prim;
            b.attributes.polygon.boundary = 0;
            b.verts[0] = prim.verts[i];
            b.verts[1] = prim.verts[(i < prim.verts.size()-1) ? i+1 : 0];
            for (int j=0; j<4; j++) {
                b.verts[0].rgba[j] = b.verts[1].rgba[j] = 0.0;
            }
            primitives.add(&b);
        }
    }
}

void GLFBspTree::buildPolygonBoundary(void)
{
    back->buildPolygonBoundary();
    for (int i=0; i<primitives.size(); i++) {
        if (primitives[i].attributes.polygon.boundary) {
            addBoundaryInList(primitives[i]);
        }
    }
    front->buildPolygonBoundary();
}
#endif
