/*
 * glFeedback, an OpenGL Feedback processing Library
 * Copyright (C) 2003 Francesco Bradascio
 *
 * E-mail: fbradasc@katamail.com
 * URL: http://digilander.libero.it/fbradasc/GL_FEEDBACK.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This work is mostly based on the GL2PS Library which is:
 *
 * Copyright (C) 1999-2003 Christophe Geuzaine
 * E-mail: geuz@geuz.org
 * URL: http://www.geuz.org/gl2ps/
 *
 * GL2PS Contributor(s):
 *   Michael Sweet <mike@easysw.com>
 *   Marc Ume <marc.ume@digitalgraphics.be>
 *   Jean-Francois Remacle <remacle@scorec.rpi.edu>
 *   Bart Kaptein <B.L.Kaptein@lumc.nl>
 *   Quy Nguyen-Dai<quy@vnilux.com>
 *   Sam Buss <sbuss@ucsd.edu>
 *   Shane Hill <Shane.Hill@dsto.defence.gov.au>
 *   Romain Boman <r_boman@yahoo.fr>
 *   Rouben Rostamian <rostamian@umbc.edu>
 *   Diego Santa Cruz <Diego.SantaCruz@epfl.ch>
 *   Shahzad Muzaffar <Shahzad.Muzaffar@cern.ch>
 *   Lassi Tuura <lassi.tuura@cern.ch>
 *   Guy Barrand <barrand@lal.in2p3.fr>
 */

#include <string.h>
#include <time.h>

#include <GLF/GLFContext.h>
#include <GLF/GLFBuffers.h>
#include <GLF/GLFBspTree.h>
#include <GLF/GLFMath.h>
#include <GLF/GLFViewport.h>

static GLFBuffers feedbackBuffers;

GLint GLFContext::getVertex(GLFVertex &v, GLfloat *p)
{
GLint i;

    v.xyz[0] = p[0];
    v.xyz[1] = p[1];
    v.xyz[2] = GLF_DEPTH_FACT * p[2];

    if (colormode == GL_COLOR_INDEX && colorsize > 0) {
        i = (GLint)(p[3] + 0.5);
        v.rgba = colormap[i];
        return 4;
    } else {
        v.rgba = &(p[3]);
        return 7;
    }
}

GLFContext::GLFContext (
    const char *title,
    const char *creator,
    const char *producer,
    const char *copyright,
    GLint colormode, GLint colorsize, GLfloat *colormap,
    GLint nr, GLint ng, GLint nb
) {
    this->first         = 0;
    this->last          = 0;
    this->viewport      = 0;
    this->title         = 0;
    this->creator       = 0;
    this->producer      = 0;
    this->copyright     = 0;
    this->colorsize     = 0;
    this->colormap      = 0;
    this->colormode     = 0;
    this->options       = 0;
    this->buffsize      = 0;
    this->feedback      = 0;
    this->maxbestroot   = 10;
    this->boundary      = 0;
    this->parseError    = GLF_SUCCESS;

    if (title) {
        this->title = strdup(title);
    }
    if (creator) {
        this->creator = strdup(creator);
    }
    if (producer) {
        this->producer = strdup(producer);
    }
    if (copyright) {
        this->copyright = strdup(copyright);
    }
    this->colormode = colormode;
    if (this->colormode == GL_COLOR_INDEX) {
        if (!colorsize || !colormap) {
            return; // Missing colormap for GL_COLOR_INDEX rendering
        }
        this->colorsize = colorsize;
        this->colormap  = new GLFrgba[this->colorsize];
        for (size_t i=0; i<this->colorsize; i++) {
            this->colormap[i] = &(colormap[i*4]);
        }
    } else if (this->colormode != GL_RGBA) {
        return; // Unknown color mode
    }
    this->threshold.set (
        (nr) ? 1.0/(GLfloat)nr : 0.032,
        (ng) ? 1.0/(GLfloat)ng : 0.017,
        (nb) ? 1.0/(GLfloat)nb : 0.05,
        1.0
    );
}

GLFContext::~GLFContext()
{
    if (title    ) { delete title    ; }
    if (creator  ) { delete creator  ; }
    if (producer ) { delete producer ; }
    if (copyright) { delete copyright; }
    if (colormap ) { delete colormap ; }
    setFeedbackBuffer(0); // free the feedback buffer
}

bool GLFContext::setFeedbackBuffer(GLsizei buffsize)
{
    this->buffsize = (buffsize>=0) ? buffsize : 0;

    if (this->buffsize>0) {
        if (!feedback) {
            feedback = feedbackBuffers.get();
        }
        feedback->resize(this->buffsize);
    } else {
        if (this->feedback) {
            feedback->used = false;
        }
        this->feedback = 0;
    }
    parseError = GLF_SUCCESS;

    GLFViewport *vp1, *vp2;
    vp1 = first;
    while (vp1) {
        vp2 = vp1->next;
        delete vp1;
        vp1 = vp2;
    }
    viewport = first = last = 0;

    return (this->buffsize && this->feedback && this->feedback->buffer);
}

bool GLFContext::active(bool skipModeCheck)
{
GLint mode;

    if (!skipModeCheck) {
        glGetIntegerv(GL_RENDER_MODE,&mode);
    } else {
        mode = GL_FEEDBACK;
    }
    return ( feedback && feedback->buffer && mode==GL_FEEDBACK ) ;
}

void GLFContext::printPrimitive(GLFPrimitive& prim, void *ctx)
{
    ((GLFContext *)ctx)->printPrimitive(prim);
}

void GLFContext::drawPixels (
    GLsizei width , GLsizei height,
    GLenum  format, GLenum  type  ,
    const void *pixels
) {
GLfloat pos[4], rgba[4];
GLboolean valid;

    if (viewport) {
        if (!pixels || width*height<=0 || format!=GL_RGB || type!=GL_FLOAT) {
            return;
        }
        glGetBooleanv(GL_CURRENT_RASTER_POSITION_VALID, &valid);
        if (!valid) { // the pixmap is culled
            return;
        }
        glGetFloatv(GL_CURRENT_RASTER_POSITION, pos );
        glGetFloatv(GL_CURRENT_RASTER_COLOR   , rgba);
        GLFVertex vertex (
            pos[0], pos[1], GLF_DEPTH_FACT*pos[2],
            rgba[0], rgba[1], rgba[2], rgba[3]
        );
        GLFPrimitive *p = new GLFPrimitive (
            GLF_PIXMAP,
            &vertex,
            attributes,
            format, type,
            width, height,
            (GLfloat *)pixels
        );
        viewport->add(p);
    }
}

void GLFContext::pointSize(GLfloat size)
{
    attributes.point.size = size;
    glPassThrough(GLF_SET_POINT_SIZE);
    glPassThrough(attributes.point.size);
}

void GLFContext::lineWidth(GLfloat width)
{
    attributes.line.width = width;
    glPassThrough(GLF_SET_LINE_WIDTH);
    glPassThrough(attributes.line.width);
}

void GLFContext::lineStipple(GLint factor, GLushort pattern)
{
    if (glIsEnabled(GL_LINE_STIPPLE)) {
        attributes.line.stippleFactor  = factor;
        attributes.line.stipplePattern = pattern;
        glPassThrough(GLF_SET_LINE_STIPPLE);
        glPassThrough (
            (GLfloat) (
                ((factor & 0xffff)<<16) | (pattern & 0xffff)
            )
        );
    } else if (
            attributes.line.stippleFactor  != 1 &&
            attributes.line.stipplePattern != 0xffff
    ) {
        attributes.line.stippleFactor  = 1;
        attributes.line.stipplePattern = 0xffff;
        glPassThrough(GLF_CLEAR_LINE_STIPPLE);
    }
}

void GLFContext::polygonOffset(GLfloat factor, GLfloat units)
{
    if (glIsEnabled(GL_POLYGON_OFFSET_FILL)) {
        if (!attributes.polygon.offset) {
            attributes.polygon.offset = new GLfloat[2];
        }
        attributes.polygon.offset[0] = factor;
        attributes.polygon.offset[1] = units;
        glPassThrough(GLF_BEGIN_POLYGON_OFFSET_FILL);
        glPassThrough(attributes.polygon.offset[0]);
        glPassThrough(attributes.polygon.offset[1]);
    } else if (attributes.polygon.offset) {
        delete attributes.polygon.offset;
        attributes.polygon.offset = 0;
        glPassThrough(GLF_END_POLYGON_OFFSET_FILL);
    }
}

void GLFContext::polygonStipple(const GLubyte *mask)
{
    if (glIsEnabled(GL_POLYGON_STIPPLE)) {
        if (!attributes.polygon.stippleMask) {
            attributes.polygon.stippleMask =
                new GLfloat[32];
        }
        memcpy(attributes.polygon.stippleMask,mask,32*sizeof(GLfloat));
        glPassThrough(GLF_SET_POLYGON_STIPPLE);
        for (int i=0; i<32; i++) {
            glPassThrough(attributes.polygon.stippleMask[i]);
        }
    } else if (attributes.polygon.stippleMask) {
        delete attributes.polygon.stippleMask;
        attributes.polygon.stippleMask = 0;
        glPassThrough(GLF_CLEAR_POLYGON_STIPPLE);
    }
}

void GLFContext::enable(GLenum cap)
{
}

void GLFContext::disable(GLenum cap)
{
    switch(cap) {
        case GL_LINE_STIPPLE:
            lineStipple(0,0);
        break;
        case GL_POLYGON_OFFSET_FILL:
            polygonOffset(0.0,0.0);
        break;
        case GL_POLYGON_STIPPLE:
            polygonStipple(0);
        break;
        default:
        break;
    }
}

GLint GLFContext::fetchPrimitives()
{
    if (parseError != GLF_OVERFLOW && viewport) {
        parseError = viewport->parse();
    }
    return parseError;
}

GLint GLFContext::printPrimitives()
{
GLFxyz eye(0.0,0.0,100000.0);

    if (parseError!=GLF_OVERFLOW && first) {
        writeFileHeader();
        for (GLFViewport *vp = first; vp; vp = vp->next) {
            writeFrameHeader(vp->viewport, vp->background);
            vp->printPrimitives(eye);
            writeFrameFooter();
        }
        writeFileFooter();
    }
    return parseError;
}

void GLFContext::beginViewport(GLint vp[4])
{
GLint index;
GLFrgba rgba;

    if (parseError != GLF_OVERFLOW && viewport) {
        if (viewport->parent) {
            //
            // close the open viweport at same level
            //
            endViewport();
        } else {
            //
            // flush the new elements for the current viewport
            //
            parseError = viewport->parse();
        }
    }
    if (parseError != GLF_OVERFLOW) {
        if (colormode == GL_RGBA || colorsize == 0) {
            GLfloat frgba[4];
            glGetFloatv(GL_COLOR_CLEAR_VALUE,frgba);
            rgba = frgba;
        } else {
            glGetIntegerv(GL_INDEX_CLEAR_VALUE, &index);
            rgba = colormap[index];
            rgba[3] = 0.0;
        }
        GLFViewport *newViewport = new GLFViewport(vp,rgba,viewport,this);
        if (!viewport) { // first call to beginViewport or new page
            if (!first) {
                first = newViewport;
            } else {
                first->next = newViewport;
                newViewport->prev = first;
            }
            last = newViewport;
        }
        viewport = newViewport;
    }
}

void GLFContext::endViewport(bool closeAll)
{
    //
    // flush the new elements for the last viewport and set as current
    // viewport the viewport's parent
    //
    if (
        viewport && parseError != GLF_OVERFLOW &&
        (parseError = viewport->parse()) != GLF_OVERFLOW
    ) {
        viewport = viewport->parent;
        if (closeAll) {
            endViewport(closeAll);
        }
    }
}
