/*
 * glFeedback, an OpenGL Feedback processing Library
 * Copyright (C) 2003 Francesco Bradascio
 *
 * E-mail: fbradasc@katamail.com
 * URL: http://digilander.libero.it/fbradasc/GL_FEEDBACK.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This work is mostly based on the GL2PS Library which is:
 *
 * Copyright (C) 1999-2003 Christophe Geuzaine
 * E-mail: geuz@geuz.org
 * URL: http://www.geuz.org/gl2ps/
 *
 * GL2PS Contributor(s):
 *   Michael Sweet <mike@easysw.com>
 *   Marc Ume <marc.ume@digitalgraphics.be>
 *   Jean-Francois Remacle <remacle@scorec.rpi.edu>
 *   Bart Kaptein <B.L.Kaptein@lumc.nl>
 *   Quy Nguyen-Dai<quy@vnilux.com>
 *   Sam Buss <sbuss@ucsd.edu>
 *   Shane Hill <Shane.Hill@dsto.defence.gov.au>
 *   Romain Boman <r_boman@yahoo.fr>
 *   Rouben Rostamian <rostamian@umbc.edu>
 *   Diego Santa Cruz <Diego.SantaCruz@epfl.ch>
 *   Shahzad Muzaffar <Shahzad.Muzaffar@cern.ch>
 *   Lassi Tuura <lassi.tuura@cern.ch>
 *   Guy Barrand <barrand@lal.in2p3.fr>
 */

#include <string.h>
#include <stdarg.h>
#include <time.h>

#include <GLF/GLFPostscript.h>

GLFPostscript::GLFPostscript (
    const char *filename        ,
    GLint       screenSize[4]   ,
    const char *title           ,
    const char *creator         ,
    const char *producer        ,
    const char *copyright       ,
    bool        enablePs3Shading,
    GLint       colormode       ,
    GLint       colorsize       ,
    GLfloat    *colormap        ,
    GLint       nr              ,
    GLint       ng              ,
    GLint       nb
) : GLFContext (
    title,
    creator,
    producer,
    copyright,
    colormode, colorsize, colormap,
    nr, ng, nb
) {
    pageLevel         = 0;
    viewportLevel     = 0;
    pages             = 0;
    lastlinewidth     = -1;
    filename_         = strdup(filename);
    stream_           = 0;
    pageWidth_        = 0.0;
    pageHeight_       = 0.0;
    borderSize_       = 0.0;
    orientation_      = AS_IS;
    umFactor_         = 1.0;
    for (int i=0; i<4; i++) {
        this->screenSize[i] = screenSize[i];
    }
    enablePs3Shading_ = enablePs3Shading;
    resetColor();
}

GLFPostscript::~GLFPostscript()
{
    writeFileFooter();
    if (filename_) {
        delete filename_;
    }
}

void GLFPostscript::pageSetup (
    GLint   orientation,
    GLfloat pageWidth,
    GLfloat pageHeight,
    GLfloat borderSize,
    GLfloat umFactor
) {
    pageWidth_   = pageWidth;
    pageHeight_  = pageHeight;
    borderSize_  = borderSize;
    orientation_ = (orientation>=0 && orientation<=2) ? orientation : AS_IS ;
    umFactor_    = (umFactor>0) ? umFactor : 1.0;  
}

GLfloat GLFPostscript::toPsPoint(GLfloat v, GLfloat umFactor)
{
    return v * ( (umFactor <= 0.0) ? umFactor_ : umFactor );
}

void GLFPostscript::resetColor()
{
    lastrgba.set(-1.0,-1.0,-1.0,-1.0);
}

int GLFPostscript::psPrint(const char *fmt, ...)
{
va_list args;
int ret;

    va_start(args,fmt);
    ret = vfprintf(stream_,fmt,args);
    va_end(args);
    return ret;
}

void GLFPostscript::writeByte(unsigned char byte)
{
unsigned char h = byte / 16;
unsigned char l = byte % 16;

    psPrint("%x%x", h, l);
}

int GLFPostscript::getRgb (
    GLfloat *pixels, GLsizei w, GLsizei h, GLuint x, GLuint y,
    GLfloat &red, GLfloat &green, GLfloat &blue
) {
GLfloat *pimag;

    // OpenGL image is from down to up, PS image is up to down

    pimag = pixels + 3 * (w * (h - 1 - y) + x);
    red   = *pimag; pimag++;
    green = *pimag; pimag++;
    blue  = *pimag; pimag++;
    return 1;
}

void GLFPostscript::writePixmap (
    GLFPrimitive& pixmap
) {
int status = 1, nbhex, nbyte2, nbyte4, nbyte8;
GLsizei row, col, col_max;
float dr, dg, db, fgrey;
unsigned char red, green, blue, b, grey;
// FIXME: this has to be generalized...
int shade = 0;
int nbit = 4;
GLfloat x, y;
GLsizei width, height;
GLenum format, type;
GLfloat *pixels;

    if ((pixmap.width <= 0) || (pixmap.height <= 0)) {
        return;
    }
    width  = pixmap.width;
    height = pixmap.height;
    x      = pixmap.verts[0].xyz[0];
    y      = pixmap.verts[0].xyz[1];
    format = pixmap.format;
    type   = pixmap.imageType;
    pixels = pixmap.pixels;
    
    psPrint("gsave\n");
    psPrint("%.2f %.2f translate\n", x, y); 
    psPrint("%d %d scale\n", width, height); 

    if (shade != 0) { /* grey */
        psPrint("/picstr %d string def\n", width); 
        psPrint("%d %d %d\n", width, height, 8); 
        psPrint("[ %d 0 0 -%d 0 %d ]\n", width, height, height); 
        psPrint("{ currentfile picstr readhexstring pop }\n");
        psPrint("image\n");
        for (row = 0; row < height; row++) {
            for (col = 0; col < width; col++) { 
                status = getRgb (
                    pixels, width, height, col, row, dr, dg, db
                ) == 0 ? 0 : status;
                fgrey  = (0.30 * dr + 0.59 * dg + 0.11 * db);
                grey   = (unsigned char)(255. * fgrey);
                writeByte(grey);
            }
            psPrint("\n");
        }
        nbhex = width * height * 2; 
        psPrint("%%%% nbhex digit          :%d\n", nbhex); 
    } else if (nbit == 2) { 
        nbyte2 = (width * 3)/4;
        nbyte2 /=3;
        nbyte2 *=3;
        col_max = (nbyte2 * 4)/3;
        /* 2 bit for r and g and b */
        /* rgbs following each other */
        psPrint("/rgbstr %d string def\n", nbyte2); 
        psPrint("%d %d %d\n", col_max, height, 2); 
        psPrint("[ %d 0 0 -%d 0 %d ]\n", col_max, height, height); 
        psPrint("{ currentfile rgbstr readhexstring pop }\n" );
        psPrint("false 3\n" );
        psPrint("colorimage\n" );
        for (row = 0; row < height; row++) {
            for (col = 0; col < col_max; col+=4) {
                status = getRgb (
                    pixels, width, height, col, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(3. * dr);
                green  = (unsigned char)(3. * dg);
                blue   = (unsigned char)(3. * db);
                b      = red;
                b      = (b<<2)+green;
                b      = (b<<2)+blue;
                status = getRgb (
                    pixels, width, height, col+1, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(3. * dr);
                green  = (unsigned char)(3. * dg);
                blue   = (unsigned char)(3. * db);
                b      = (b<<2)+red;
                writeByte(b);
    
                b      = green;
                b      = (b<<2)+blue;
                status = getRgb (
                    pixels, width, height, col+2, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(3. * dr);
                green  = (unsigned char)(3. * dg);
                blue   = (unsigned char)(3. * db);
                b      = (b<<2)+red;
                b      = (b<<2)+green;
                writeByte(b);
    
                b      = blue;
                status = getRgb (
                    pixels,width,height, col+3, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(3. * dr);
                green  = (unsigned char)(3. * dg);
                blue   = (unsigned char)(3. * db);
                b      = (b<<2)+red;
                b      = (b<<2)+green;
                b      = (b<<2)+blue;
                writeByte(b);
            }
            psPrint("\n");
        }
    } else if (nbit == 4) { 
        nbyte4 = (width  * 3)/2;
        nbyte4 /=3;
        nbyte4 *=3;
        col_max = (nbyte4 * 2)/3;
        /* 4 bit for r and g and b */
        /* rgbs following each other */
        psPrint("/rgbstr %d string def\n", nbyte4);
        psPrint("%d %d %d\n", col_max, height,4);
        psPrint("[ %d 0 0 -%d 0 %d ]\n", col_max, height, height);
        psPrint("{ currentfile rgbstr readhexstring pop }\n");
        psPrint("false 3\n");
        psPrint("colorimage\n");
        for (row = 0; row < height; row++) {
            for (col = 0; col < col_max; col+=2) {
                status = getRgb (
                    pixels, width, height, col, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(15. * dr);
                green  = (unsigned char)(15. * dg);
                psPrint("%x%x", red, green);
                blue   = (unsigned char)(15. * db);
        
                status = getRgb (
                    pixels, width, height, col+1, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(15. * dr);
                psPrint("%x%x",blue,red);
                green  = (unsigned char)(15. * dg);
                blue   = (unsigned char)(15. * db);
                psPrint("%x%x", green, blue);
            }
            psPrint("\n");
        }
    } else{ 
        nbyte8 = width * 3;
        /* 8 bit for r and g and b */
        psPrint("/rgbstr %d string def\n", nbyte8);
        psPrint("%d %d %d\n", width, height, 8);
        psPrint("[ %d 0 0 -%d 0 %d ]\n", width, height, height); 
        psPrint("{ currentfile rgbstr readhexstring pop }\n");
        psPrint("false 3\n");
        psPrint("colorimage\n");
        for (row = 0; row < height; row++) {
            for (col = 0; col < width; col++) {
                status = getRgb (
                    pixels, width, height, col, row, dr, dg, db
                ) == 0 ? 0 : status;
                red    = (unsigned char)(255. * dr);
                writeByte(red);
                green  = (unsigned char)(255. * dg);
                writeByte(green);
                blue   = (unsigned char)(255. * db);
                writeByte(blue);
            }
            psPrint("\n");
        }
    }
    if (status == 0) {
        // "Problem to retrieve some pixel rgb"
    }
    psPrint("grestore\n");
}

void GLFPostscript::writeFileHeader()
{
time_t now;
GLfloat x, y, w, h;


    if (!feedback || !feedback->buffer) {
        return;
    }
    stream_ = fopen(filename_,"w+");

    time(&now);

    psPrint("%%!PS-Adobe-3.0\n");
    psPrint (
        "%%%%Title: %s\n"
        "%%%%Creator: %s\n"
        "%%%%For: %s\n"
        "%%%%CreationDate: %s"
        "%%%%LanguageLevel: 3\n"
        "%%%%DocumentData: Clean7Bit\n"
        "%%%%Orientation: %s\n",
        title, creator, producer, ctime(&now),
        (orientation_ == LANDSCAPE) ? "Landscape" : "Portrait"
    );
    if (orientation_!=AS_IS) {
        x = y = 0;
        w = toPsPoint(pageWidth_);
        h = toPsPoint(pageHeight_);
    } else {
        x = screenSize[0];
        y = screenSize[1];
        w = screenSize[2];
        h = screenSize[3];
    }
    psPrint (
        "%%%%Pages: (atend)\n"
        "%%%%PageOrder: Ascend\n"
        "%%%%BoundingBox: %g %g %g %g\n"
        "%%%%Copyright: %s\n"
        "%%%%EndComments\n",
        x, y, w, h,
        copyright
    );
    /*
    RGB color: r g b C (replace C by G in output to change from rgb to gray)
    Grayscale: r g b G
    Font choose: size fontname FC
    String primitive: (string) x y size fontname S
    Point primitive: x y size P
    Line width: width W
    Flat-shaded line: x2 y2 x1 y1 L
    Smooth-shaded line: x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 SL
    Flat-shaded triangle: x3 y3 x2 y2 x1 y1 T
    Smooth-shaded triangle: x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 ST
    */
    psPrint (
        "%%%%BeginProlog\n"
        "/glfdict 64 dict def glfdict begin\n"
        "1 setlinecap 1 setlinejoin\n"
        "/tryPS3shading %s def %% set to false to force subdivision\n"
        "/rThreshold %g def %% red component subdivision threshold\n"
        "/gThreshold %g def %% green component subdivision threshold\n"
        "/bThreshold %g def %% blue component subdivision threshold\n"
        "/BD { bind def } bind def\n"
        "/C  { setrgbcolor } BD\n"
        "/G  { 0.082 mul exch 0.6094 mul add exch 0.3086 mul add neg 1.0 add setgray } BD\n"
        "/W  { setlinewidth } BD\n"
        "/FC { findfont exch scalefont setfont } BD\n"
        "/S  { FC moveto show } BD\n"
        "/P  { newpath 0.0 360.0 arc closepath fill } BD\n"
        "/L  { newpath moveto lineto stroke } BD\n"
        "/SL { C moveto C lineto stroke } BD\n"
        "/T  { newpath moveto lineto lineto closepath fill } BD\n",
        (enablePs3Shading_) ? "true" : "false",
        threshold[0], threshold[1], threshold[2]
    );
    /*
    Smooth-shaded triangle with PostScript level 3 shfill operator:
    x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 STshfill
    */
    psPrint (
        "/STshfill {\n"
        "  /b1 exch def /g1 exch def /r1 exch def /y1 exch def /x1 exch def\n"
        "  /b2 exch def /g2 exch def /r2 exch def /y2 exch def /x2 exch def\n"
        "  /b3 exch def /g3 exch def /r3 exch def /y3 exch def /x3 exch def\n"
        "  gsave << /ShadingType 4 /ColorSpace [/DeviceRGB]\n"
        "  /DataSource [ 0 x1 y1 r1 g1 b1 0 x2 y2 r2 g2 b2 0 x3 y3 r3 g3 b3 ] >>\n"
        "  shfill grestore } BD\n"
    );
    /*
    Flat-shaded triangle with middle color:
    x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 Tm
    */
    psPrint (
        /* stack : x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 */
        /* r = (r1+r2+r3)/3 */
        "/Tm { 3 -1 roll 8 -1 roll 13 -1 roll add add 3 div\n"
        /* stack : x3 y3 g3 b3 x2 y2 g2 b2 x1 y1 g1 b1 r */
        /* g = (g1+g2+g3)/3 */
        "      3 -1 roll 7 -1 roll 11 -1 roll add add 3 div\n"
        /* stack : x3 y3 b3 x2 y2 b2 x1 y1 b1 r g b */
        /* b = (b1+b2+b3)/3 */
        "      3 -1 roll 6 -1 roll 9 -1 roll add add 3 div"
        /* stack : x3 y3 x2 y2 x1 y1 r g b */
        " C T } BD\n"
    );
    /*
    Split triangle in four sub-triangles (at sides middle points) and call
    the STnoshfill procedure on each, interpolating the colors in RGB space:
          x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 STsplit
       (in procedure comments key: (Vi) = xi yi ri gi bi)
    */
    psPrint (
        "/STsplit {\n"
        "      4 index 15 index add 0.5 mul\n" /* x13 = (x1+x3)/2 */
        "      4 index 15 index add 0.5 mul\n" /* y13 = (y1+y3)/2 */
        "      4 index 15 index add 0.5 mul\n" /* r13 = (r1+r3)/2 */
        "      4 index 15 index add 0.5 mul\n" /* g13 = (g1+g3)/2 */
        "      4 index 15 index add 0.5 mul\n" /* b13 = (b1+b3)/2 */
        "      5 copy 5 copy 25 15 roll\n"
        /* stack : (V3) (V13) (V13) (V13) (V2) (V1) */
        "      9 index 30 index add 0.5 mul\n" /* x23 = (x2+x3)/2 */
        "      9 index 30 index add 0.5 mul\n" /* y23 = (y2+y3)/2 */
        "      9 index 30 index add 0.5 mul\n" /* r23 = (r2+r3)/2 */
        "      9 index 30 index add 0.5 mul\n" /* g23 = (g2+g3)/2 */
        "      9 index 30 index add 0.5 mul\n" /* b23 = (b2+b3)/2 */
        "      5 copy 5 copy 35 5 roll 25 5 roll 15 5 roll\n"
        /* stack : (V3) (V13) (V23) (V13) (V23) (V13) (V23) (V2) (V1) */
        "      4 index 10 index add 0.5 mul\n" /* x12 = (x1+x2)/2 */
        "      4 index 10 index add 0.5 mul\n" /* y12 = (y1+y2)/2 */
        "      4 index 10 index add 0.5 mul\n" /* r12 = (r1+r2)/2 */
        "      4 index 10 index add 0.5 mul\n" /* g12 = (g1+g2)/2 */
        "      4 index 10 index add 0.5 mul\n" /* b12 = (b1+b2)/2 */
        "      5 copy 5 copy 40 5 roll 25 5 roll 15 5 roll 25 5 roll\n"
        /* stack : (V3) (V13) (V23) (V13) (V12) (V23) (V13) (V1) (V12)
                   (V23) (V12) (V2)
        */
        "      STnoshfill STnoshfill STnoshfill STnoshfill } BD\n"
    );
    /*
     Gouraud shaded triangle using recursive subdivision until the
     difference between corner colors does not exceed the thresholds:
          x3 y3 r3 g3 b3 x2 y2 r2 g2 b2 x1 y1 r1 g1 b1 STnoshfill
    */
    psPrint (
        "/STnoshfill {\n"
        /* |r1-r2|>rth */
        "      2 index 8 index sub abs rThreshold gt\n"
        "      { STsplit }\n"
        /* |g1-g2|>gth */
        "      { 1 index 7 index sub abs gThreshold gt\n"
        "        { STsplit }\n"
        /* |b1-b2|>bth */
        "        { dup 6 index sub abs bThreshold gt\n"
        "          { STsplit }\n"
        /* |r1-r3|>rht */
        "          { 2 index 13 index sub abs rThreshold gt\n"
        "            { STsplit }\n"
        /* |g1-g3|>gth */
        "            { 1 index 12 index sub abs gThreshold gt\n"
        "              { STsplit }\n"
        /* |b1-b3|>bth */
        "              { dup 11 index sub abs bThreshold gt\n"
        "                { STsplit }\n"
        /* |r2-r3|>rht */
        "                { 7 index 13 index sub abs rThreshold gt\n"
        "                  { STsplit }\n"
        /* |g2-g3|>gth */
        "                  { 6 index 12 index sub abs gThreshold gt\n"
        "                    { STsplit }\n"
        /* |b2-b3|>bth */
        "                    { 5 index 11 index sub abs bThreshold gt\n"
        "                      { STsplit }\n"
        /* all colors sufficiently similar */
        "                      { Tm }\n"
        "                      ifelse }\n"
        "                    ifelse }\n"
        "                  ifelse }\n"
        "                ifelse }\n"
        "              ifelse }\n"
        "            ifelse }\n"
        "          ifelse }\n"
        "        ifelse }\n"
        "      ifelse } BD\n"
    );
    psPrint (
        "tryPS3shading\n"
        "{ /shfill where\n"
        "  { /ST { STshfill } BD }\n"
        "  { /ST { STnoshfill } BD }\n"
        "  ifelse }\n"
        "{ /ST { STnoshfill } BD }\n"
        "ifelse\n"
    );
    psPrint (
        "end\n"
        "%%%%EndProlog\n"
        "%%%%BeginSetup\n"
        "/DeviceRGB setcolorspace\n"
        "glfdict begin\n"
        "%%%%EndSetup\n"
    );
}

void GLFPostscript::writeFrameHeader(GLint viewport[4], GLFrgba& background)
{
    pages++;
    psPrint (
        "%%%%Page: %d %d\n"
        "%%%%BeginPageSetup\n",
        pages, pages
    );
    if (orientation_!=AS_IS) {
        GLfloat tx, ty, wp, hp, wi, hi, sx, sy, scale;
        tx = ty = toPsPoint(borderSize_);
        wi = viewport[2];
        hi = viewport[3];
        if (orientation_ == LANDSCAPE) {
            wp = toPsPoint(pageHeight_) - 2*tx;
            hp = toPsPoint(pageWidth_) - 2*ty;
            psPrint (
                "%d 0 translate 90 rotate\n",
                (int)wp
            );
            ty = (wp - hp) - ty;
        } else {
            wp = toPsPoint(pageWidth_) - 2*tx;
            hp = toPsPoint(pageHeight_) - 2*ty;
        }
        sx = wp/wi;
        sy = hp/hi;
        scale = (sx < sy) ? sx : sy;
        tx += (wp-(wi*scale))/2.0;
        ty += (hp-(hi*scale))/2.0;
        tx = (tx > 0.0 && tx < 0.5) ? 0.0 : tx;
        ty = (ty > 0.0 && ty < 0.5) ? 0.0 : ty;
        psPrint (
            "%g %g translate\n"
            "%g %g scale\n"
            "%%%%EndPageSetup\n"
            "mark\n"
            "gsave\n",
            tx, ty,
            scale, scale
        );
    } else {
        psPrint (
            "%%%%EndPageSetup\n"
            "mark\n"
            "gsave\n"
            "1.0 1.0 scale\n"
        );
    }
    if (options & GLF_OPTION_FLAG(GLF_DRAW_BACKGROUND)) {
        psPrint (
            "%g %g %g C\n"
            "newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto\n"
            "closepath fill\n",
            background[0], background[1], background[2], 
            viewport[0], viewport[1], viewport[2], 
            viewport[1], viewport[2], viewport[3], 
            viewport[0], viewport[3]
        );
    }
    pageLevel++;
}

void GLFPostscript::writeViewportHeader(GLint viewport[4], GLFrgba& background)
{
    psPrint (
        "gsave\n"
        "1.0 1.0 scale\n"
    );
    if (options & GLF_OPTION_FLAG(GLF_DRAW_BACKGROUND)) {
        psPrint (
            "%g %g %g C\n"
            "newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto\n"
            "closepath fill\n",
            background[0], background[1], background[2],
            viewport[0]            , viewport[1],
            viewport[0]+viewport[2], viewport[1],
            viewport[0]+viewport[2], viewport[1]+viewport[3],
            viewport[0]            , viewport[1]+viewport[3]
        );
        psPrint (
            "%g %g %g C\n"
            "newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto\n"
            "closepath clip\n",
            background[0], background[1], background[2],
            viewport[0]            , viewport[1],
            viewport[0]+viewport[2], viewport[1],
            viewport[0]+viewport[2], viewport[1]+viewport[3],
            viewport[0]            , viewport[1]+viewport[3]
        );
    }
    viewportLevel++;
}

void GLFPostscript::writeColor(GLFrgba rgba)
{
    if (lastrgba != rgba) {
        lastrgba = rgba;
        psPrint("%g %g %g C\n", rgba[0], rgba[1], rgba[2]);
    }
}

void GLFPostscript::printPrimitive(GLFPrimitive& prim)
{
    if ((options && GLF_OPTION_FLAG(GLF_OCCLUSION_CULL)) && prim.culled) {
        return;
    }
    switch(prim.type) {
        case GLF_PIXMAP:
            writePixmap(prim);
        break;
        case GLF_POINT:
            writeColor(prim.verts[0].rgba);
            psPrint (
                "%g %g %g P\n",
                prim.verts[0].xyz[0],
                prim.verts[0].xyz[1],
                0.5 * prim.attributes.point.size
            );
        break;
        case GLF_LINE:
            if (lastlinewidth != prim.attributes.line.width) {
                lastlinewidth = prim.attributes.line.width;
                psPrint("%g W\n", lastlinewidth);
            }
            if (prim.attributes.line.stipplePattern != 0xffff) {
                GLint pos   = 15;
                bool  onoff = true;
                bool  bit;
                GLint count;
                psPrint("[");
                while (pos>=0) {
                    pos++;
                    count = -1;
                    do {
                        count++; pos--;
                        if (pos<0) {
                            break;
                        }
                        bit = ((1<<pos) & prim.attributes.line.stipplePattern);
                        bit = (bit==0) ? false : true;
                    } while (bit==onoff);
                    count *= prim.attributes.line.stippleFactor;
                    psPrint(" %d", count);
                    onoff = !onoff;
                }
                if (!onoff) {
                    psPrint(" 0] 0 setdash\n");
                } else {
                    psPrint("] 0 setdash\n");
                }
            }
            if (prim.sameVertexColor()) {
                resetColor();
                psPrint (
                    "%g %g %g %g %g %g %g %g %g %g SL\n",
                    prim.verts[1].xyz[0], prim.verts[1].xyz[1], // x1, y1
                    prim.verts[1].rgba[0],                      // r1
                    prim.verts[1].rgba[1],                      // g1
                    prim.verts[1].rgba[2],                      // b1
                    prim.verts[0].xyz[0], prim.verts[0].xyz[1], // x0, y0
                    prim.verts[0].rgba[0],                      // r0
                    prim.verts[0].rgba[1],                      // g0
                    prim.verts[0].rgba[2]                       // b0
                );
            } else {
                writeColor(prim.verts[0].rgba);
                psPrint (
                    "%g %g %g %g L\n",
                    prim.verts[1].xyz[0], prim.verts[1].xyz[1],
                    prim.verts[0].xyz[0], prim.verts[0].xyz[1]
                );
            }
            if (prim.attributes.line.stipplePattern != 0xffff) {
                psPrint("[] 0 setdash\n");
            }
        break;
        case GLF_TRIANGLE:
            if (!prim.sameVertexColor()) {
                resetColor();
                psPrint (
                    "%g %g %g %g %g %g %g %g %g %g %g %g %g %g %g ST\n",
                    prim.verts[2].xyz[0], prim.verts[2].xyz[1], // x2, y2
                    prim.verts[2].rgba[0],                      // r2
                    prim.verts[2].rgba[1],                      // g2
                    prim.verts[2].rgba[2],                      // b2
                    prim.verts[1].xyz[0], prim.verts[1].xyz[1], // x1, y1
                    prim.verts[1].rgba[0],                      // r1
                    prim.verts[1].rgba[1],                      // g1
                    prim.verts[1].rgba[2],                      // b1
                    prim.verts[0].xyz[0], prim.verts[0].xyz[1], // x0, y0
                    prim.verts[0].rgba[0],                      // r0
                    prim.verts[0].rgba[1],                      // g0
                    prim.verts[0].rgba[2]                       // b0
                );
            } else{
                writeColor(prim.verts[0].rgba);
                psPrint (
                    "%g %g %g %g %g %g T\n",
                    prim.verts[2].xyz[0], prim.verts[2].xyz[1],
                    prim.verts[1].xyz[0], prim.verts[1].xyz[1],
                    prim.verts[0].xyz[0], prim.verts[0].xyz[1]
                );
            }
        break;
        case GLF_QUADRANGLE:
            // There should not be any quad left to print
        break;
        default:
            // Unknown type of primitive to print
        break;
    }
}

void GLFPostscript::writeFrameFooter()
{
    pageLevel--;
    psPrint (
        "grestore\n"
        "showpage\n"
        "cleartomark\n"
        "%%%%PageTrailer\n"
    );
}

void GLFPostscript::writeViewportFooter()
{
    viewportLevel--;
    psPrint (
        "grestore\n"
    );
}

void GLFPostscript::writeFileFooter()
{
    if (stream_) {
        while (viewportLevel>0) {
            writeViewportFooter();
        }
        while (pageLevel>0) {
            writeFrameFooter();
        }
        psPrint (
            "%%%%Trailer\n"
            "end\n"
            "%%%%Pages: %d\n"
            "%%%%EOF\n",
            pages
        );
        fflush(stream_);
        fclose(stream_);
    }
    stream_ = 0;
}

void GLFPostscript::flush()
{
    writeFileFooter();
}
