/*
 * glFeedback, an OpenGL Feedback processing Library
 * Copyright (C) 2003 Francesco Bradascio
 *
 * E-mail: fbradasc@katamail.com
 * URL: http://digilander.libero.it/fbradasc/GL_FEEDBACK.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This work is mostly based on the GL2PS Library which is:
 *
 * Copyright (C) 1999-2003 Christophe Geuzaine
 * E-mail: geuz@geuz.org
 * URL: http://www.geuz.org/gl2ps/
 *
 * GL2PS Contributor(s):
 *   Michael Sweet <mike@easysw.com>
 *   Marc Ume <marc.ume@digitalgraphics.be>
 *   Jean-Francois Remacle <remacle@scorec.rpi.edu>
 *   Bart Kaptein <B.L.Kaptein@lumc.nl>
 *   Quy Nguyen-Dai<quy@vnilux.com>
 *   Sam Buss <sbuss@ucsd.edu>
 *   Shane Hill <Shane.Hill@dsto.defence.gov.au>
 *   Romain Boman <r_boman@yahoo.fr>
 *   Rouben Rostamian <rostamian@umbc.edu>
 *   Diego Santa Cruz <Diego.SantaCruz@epfl.ch>
 *   Shahzad Muzaffar <Shahzad.Muzaffar@cern.ch>
 *   Lassi Tuura <lassi.tuura@cern.ch>
 *   Guy Barrand <barrand@lal.in2p3.fr>
 */

#include <stdlib.h>

#include <GLF/GLFPrimitives.h>

GLFPrimitives::GLFPrimitives(GLFPrimitives *storage)
{
    vect_     = 0;
    size_     = 0;
    capacity_ = 0;
    storage_  = storage;
    resize(0);
}

GLFPrimitives::GLFPrimitives(const GLFPrimitives& p)
{
    vect_     = 0;
    size_     = 0;
    capacity_ = 0;
    (GLFPrimitives&)*this = p;
}

GLFPrimitives::~GLFPrimitives()
{
    if (vect_) {
        if (!storage_) {
            for (unsigned int i=0; i<size_; i++) {
                if (vect_[i]) {
                    delete vect_[i];
                }
            }
        }
        delete vect_;
    }
}

GLFPrimitives& GLFPrimitives::operator=(const GLFPrimitives& p)
{
    resize(0);
    storage_ = p.storage_;
    for (unsigned int i=0; i<p.size(); i++) {
        push_back(p[i]);
    }
    return *this;
}

void GLFPrimitives::traverse (
    void (*cb)(GLFPrimitive& prim, void *data),
    void *data,
    GLboolean forward
) {
    if (forward) {
        for (unsigned int i = 0; i<size_; i++) {
            (*cb)(*(vect_[i]),data);
        }
    } else {
        for (int i = size_-1; i>=0; i--) {
            (*cb)(*(vect_[i]),data);
        }
    }
}

GLFPrimitive& GLFPrimitives::findRoot(unsigned int maxBestRoot)
{
GLint count, best = 1000000;
GLFplane plane;
unsigned int maxp;
unsigned int root, prim1, prim2;

    root = 0;
    if (maxBestRoot>0) {
        maxp = (size_>maxBestRoot) ? maxBestRoot : size_;
        for (prim1=0; prim1<maxp; prim1++) {
            plane = vect_[prim1]->plane();
            count = 0;
            for (prim2=0; prim2<size_; prim2++) {
                if (
                    (prim2 != prim1) && 
                    (vect_[prim2]->check(plane) == GLF_SPANNING)
                ) {
                    count++;
                }
                if (count>best) {
                    break;
                }
            }
            if (count<best) {
                best = count;
                root = prim1;
                if (!count) {
                    break;
                }
            }
        }
    }
    return *(vect_[root]);
}

void GLFPrimitives::add(GLFPrimitive &p)
{
GLFPrimitive *t1=0, *t2=0;

    if (p.type != GLF_QUADRANGLE) {
        push_back(p);
    } else if ((t1 = p.t1()) && (t2 = p.t2())) {
        if (storage_) {
            t1 = storage_->push_back(*t1);
            t2 = storage_->push_back(*t2);
        }
        push_back(*t1);
        push_back(*t2);
    }
}

void GLFPrimitives::add(GLFPrimitive *p)
{
GLFPrimitive *t1=0, *t2=0;

    if (p) {
        if (p->type != GLF_QUADRANGLE) {
            t1 = p;
            if (storage_) {
                t1 = storage_->push_back(*t1);
            }
            push_back(*t1);
        } else if ((t1 = p->t1()) && (t2 = p->t2())) {
            if (storage_) {
                t1 = storage_->push_back(*t1);
                t2 = storage_->push_back(*t2);
            }
            push_back(*t1);
            push_back(*t2);
        }
    }
}

size_t GLFPrimitives::size() const
{
    return size_;
}

GLFPrimitive& GLFPrimitives::operator [] (size_t i)
{
    // return *(vect_[(i<0 || i>=size_) ? 0 : i]);
    return *(vect_[i]);
}

GLFPrimitive& GLFPrimitives::operator [] (size_t i) const
{
    return *(vect_[i]);
}

void GLFPrimitives::resize(size_t newsize)
{
size_t cap = (newsize * 9 / 64 + 1) * 8;

    if (newsize || vect_) {
        if (cap != capacity_) {
            capacity_ = cap;
            //
            // if vect_ == NULL --> realoc == malloc
            //
            vect_ = (GLFPrimitive **)realloc (
                vect_, capacity_ * sizeof(GLFPrimitive *)
            );
        }
    }
    size_ = newsize;
}

GLFPrimitive *GLFPrimitives::push_back(const GLFPrimitive& p)
{
    if (size_ == capacity_) {
        resize(size_ + 1);
    } else {
        size_++;
    }
    vect_[size_-1] = (storage_) ? (GLFPrimitive *)&p : new GLFPrimitive(p);

    return vect_[size_-1];
}

void GLFPrimitives::clear()
{
    resize(0);
}

void GLFPrimitives::qsort (
    int start,
    int end,
    int (*fcmp)(const GLFPrimitive&, const GLFPrimitive&)
) {
GLFPrimitive *pt;
bool change=false;
int p_start;
int p_end;

    if (start<end) {
        p_start = start;
        p_end   = end;
        while (p_start<p_end) {
            if (fcmp(*(vect_[p_start]),*(vect_[p_end]))) { 
                pt = vect_[p_start];
                vect_[p_start] = vect_[p_end];
                vect_[p_end] = pt;
                if (change) {
                    change  = false;
                    p_end--;
                } else {
                    change  = true;
                    p_start++;
                }
            } else {
                if (change) {
                    p_start++;
                } else {
                    p_end--;
                }
            }
        }
        this->qsort(start,p_start-1,fcmp);
        this->qsort(p_start+1,end,fcmp);
    }
}

void GLFPrimitives::sort(int (*fcmp)(const GLFPrimitive&, const GLFPrimitive&))
{
    if (size_>1 && fcmp) {
        this->qsort(0,size_-1,fcmp);
    }
}

bool GLFPrimitives::empty()
{
    return (size_==0);
}
