/*
 * glFeedback, an OpenGL Feedback processing Library
 * Copyright (C) 2003 Francesco Bradascio
 *
 * E-mail: fbradasc@katamail.com
 * URL: http://digilander.libero.it/fbradasc/GL_FEEDBACK.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This work is mostly based on the GL2PS Library which is:
 *
 * Copyright (C) 1999-2003 Christophe Geuzaine
 * E-mail: geuz@geuz.org
 * URL: http://www.geuz.org/gl2ps/
 *
 * GL2PS Contributor(s):
 *   Michael Sweet <mike@easysw.com>
 *   Marc Ume <marc.ume@digitalgraphics.be>
 *   Jean-Francois Remacle <remacle@scorec.rpi.edu>
 *   Bart Kaptein <B.L.Kaptein@lumc.nl>
 *   Quy Nguyen-Dai<quy@vnilux.com>
 *   Sam Buss <sbuss@ucsd.edu>
 *   Shane Hill <Shane.Hill@dsto.defence.gov.au>
 *   Romain Boman <r_boman@yahoo.fr>
 *   Rouben Rostamian <rostamian@umbc.edu>
 *   Diego Santa Cruz <Diego.SantaCruz@epfl.ch>
 *   Shahzad Muzaffar <Shahzad.Muzaffar@cern.ch>
 *   Lassi Tuura <lassi.tuura@cern.ch>
 *   Guy Barrand <barrand@lal.in2p3.fr>
 */

#include <GLF/GLFViewport.h>
#include <GLF/GLFMath.h>
#include <GLF/GLFBspTree.h>

GLFViewport::GLFViewport (
    GLint       *viewport,
    GLFrgba&     background,
    GLFViewport *parent,
    GLFContext  *context
) {
    for (int i=0; i<4; i++) {
        this->viewport[i] = viewport[i];
    }
    this->feedback   = 0;
    this->used       = 0;
    this->bspTree    = 0;
    this->imageTree  = 0;
    this->first      = 0;
    this->last       = 0;
    this->prev       = 0;
    this->next       = 0;
    this->parent     = parent;
    this->ctx        = context;
    this->options    = (context) ? context->options : 0;
    this->background = background;
    if (parent) {
        if (!parent->first) {
            parent->first = this;
        } else {
            parent->last->next = this;
            this->prev         = parent->last;
        }
        parent->last = this;
    }
}

GLFViewport::~GLFViewport()
{
    if (bspTree) {
        delete bspTree;
    }
    if (imageTree) {
        delete imageTree;
    }
    GLFViewport *vp2;
    GLFViewport *vp1=first;
    while (vp1) {
        vp2 = vp1->next;
        delete vp1;
        vp1 = vp2;
    }
    if (next) { next->prev = prev; }
    if (prev) { prev->next = next; }
    next = prev = first = last = parent = 0;
}

GLfloat *GLFViewport::getNextToken(int skip)
{
GLfloat *cur = feedback;

    feedback += skip;
    used     -= skip;

    return cur;
}

GLenum GLFViewport::parse()
{
GLint count, v, vtot, mode, numprim, j;
GLFVertex vertices[3];
char flag;

    feedback = ctx->feedback->buffer;

    ctx->attributes = ctx->origAttributes;

    glGetIntegerv(GL_RENDER_MODE,&mode);

    numprim = used = glRenderMode(GL_RENDER);
    if (numprim<0) {
        return GLF_OVERFLOW;
    }
    while (used>0) {
        if (ctx->attributes.polygon.boundary) {
            ctx->boundary = true;
        }
        switch ((GLint)*getNextToken()) {
            case GL_POINT_TOKEN:
                getNextToken(ctx->getVertex(vertices[0],feedback));
                add(new GLFPrimitive(GLF_POINT,vertices,ctx->attributes));
            break;
            case GL_LINE_TOKEN:
            case GL_LINE_RESET_TOKEN:
                getNextToken(ctx->getVertex(vertices[0],feedback));
                getNextToken(ctx->getVertex(vertices[1],feedback));
                add(new GLFPrimitive(GLF_LINE,vertices,ctx->attributes));
            break;
            case GL_POLYGON_TOKEN:
                count = (GLint)*getNextToken();
                v = vtot = 0;
                while (count>0 && used>0) {
                    getNextToken(ctx->getVertex(vertices[v],feedback));
                    count--; vtot++;
                    if (v==2) {
                        flag = 0;
                        if (ctx->attributes.polygon.boundary) {
                            flag = (!count && vtot == 2) ? 1|2|4
                                 : (!count)              ? 2|4
                                 : (vtot == 2)           ? 1|2 : 2;
                        }
                        add (
                            new GLFPrimitive (
                                GLF_TRIANGLE,
                                vertices,
                                ctx->attributes
                            )
                        );
                        vertices[1] = vertices[2];
                    } else {
                        v++;
                    }
                }
            break;
            case GL_BITMAP_TOKEN:
            case GL_DRAW_PIXEL_TOKEN:
            case GL_COPY_PIXEL_TOKEN:
                getNextToken(ctx->getVertex(vertices[0],feedback));
            break;
            case GL_PASS_THROUGH_TOKEN:
                switch ((GLint)*getNextToken()) {
                    case GLF_SET_POINT_SIZE:
                        getNextToken();
                        ctx->attributes.point.size = *getNextToken();
                    break;
                    case GLF_SET_LINE_WIDTH:
                        getNextToken();
                        ctx->attributes.line.width = *getNextToken();
                    break;
                    case GLF_SET_LINE_STIPPLE: {
                        getNextToken();
                        int stipple = (GLint)*getNextToken();
                        ctx->attributes.line.stippleFactor =
                            (stipple>>16) & 0xffff;
                        ctx->attributes.line.stipplePattern =
                            stipple & 0xffff;
                    } break;
                    case GLF_CLEAR_LINE_STIPPLE:
                        ctx->attributes.line.stippleFactor  = 1;
                        ctx->attributes.line.stipplePattern = 0xffff;
                    break;
                    case GLF_BEGIN_POLYGON_OFFSET_FILL:
                        if (!ctx->attributes.polygon.offset) {
                            ctx->attributes.polygon.offset = new GLfloat[2];
                        }
                        getNextToken();
                        ctx->attributes.polygon.offset[0] = *getNextToken();
                        getNextToken();
                        ctx->attributes.polygon.offset[1] = *getNextToken();
                    break;
                    case GLF_END_POLYGON_OFFSET_FILL:
                        if (ctx->attributes.polygon.offset) {
                            delete ctx->attributes.polygon.offset;
                        }
                        ctx->attributes.polygon.offset = 0;
                    break;
                    case GLF_BEGIN_POLYGON_BOUNDARY:
                        ctx->attributes.polygon.boundary = 1;
                    break;
                    case GLF_END_POLYGON_BOUNDARY:
                        ctx->attributes.polygon.boundary = 0;
                    break;
                    case GLF_SET_POLYGON_STIPPLE:
                        if (!ctx->attributes.polygon.stippleMask) {
                            ctx->attributes.polygon.stippleMask =
                                new GLfloat[32];
                        }
                        for (j=0; j<32; j++) {
                            getNextToken();
                            ctx->attributes.polygon.stippleMask[j] = 
                                *getNextToken();
                        }
                    break;
                    case GLF_CLEAR_POLYGON_STIPPLE:
                        if (ctx->attributes.polygon.stippleMask) {
                            delete ctx->attributes.polygon.stippleMask;
                        }
                        ctx->attributes.polygon.stippleMask = 0;
                    break;
                }
            break;
            default:
                // throw warning "Unknown token in buffer"
                getNextToken();
            break;
        }
    }
    if (mode == GL_FEEDBACK) {
        glRenderMode(GL_FEEDBACK);
    }
    return (numprim==0) ? GLF_NO_FEEDBACK : GLF_SUCCESS;
}

void GLFViewport::add(GLFPrimitive *prim)
{
GLfloat factor, units, area, dZ, dZdX, dZdY, maxdZ;

    if (options & GLF_OPTION_FLAG(GLF_SIMPLE_LINE_OFFSET)) {
        if (prim->type == GLF_LINE) {
            prim->verts[0].xyz[2] -= GLF_SIMPLE_OFFSET;
            prim->verts[1].xyz[2] -= GLF_SIMPLE_OFFSET;
        }
    } else if (prim->attributes.polygon.offset && prim->type == GLF_TRIANGLE) {
        //
        // FIXME: this needs some more work...
        //
        factor = prim->attributes.polygon.offset[0] / 800.0;
        units  = prim->attributes.polygon.offset[1] / 800.0;
        area = (prim->verts[1].xyz[0] - prim->verts[0].xyz[0]) *
               (prim->verts[2].xyz[1] - prim->verts[1].xyz[1]) -
               (prim->verts[2].xyz[0] - prim->verts[1].xyz[0]) *
               (prim->verts[1].xyz[1] - prim->verts[0].xyz[1]);

        dZdX = (prim->verts[2].xyz[1] - prim->verts[1].xyz[1]) *
               (prim->verts[1].xyz[2] - prim->verts[0].xyz[2]) -
               (prim->verts[1].xyz[1] - prim->verts[0].xyz[1]) *
               (prim->verts[2].xyz[2] - prim->verts[1].xyz[2]) / area;

        dZdY = (prim->verts[1].xyz[0] - prim->verts[0].xyz[0]) *
               (prim->verts[2].xyz[2] - prim->verts[1].xyz[2]) -
               (prim->verts[2].xyz[0] - prim->verts[1].xyz[0]) *
               (prim->verts[1].xyz[2] - prim->verts[0].xyz[2]) / area;

        maxdZ = sqrt(dZdX*dZdX + dZdY*dZdY);

        dZ = factor * maxdZ + units;

        prim->verts[0].xyz[2] += dZ;
        prim->verts[1].xyz[2] += dZ;
        prim->verts[2].xyz[2] += dZ;
    }
    primitives.add(*prim);
    delete prim;
}

void GLFViewport::process(GLFxyz& eye)
{
    if (bspTree) {
        delete bspTree;
    }
    if (primitives.size()) {
        bspTree = new GLFBspTree (
            ctx->maxbestroot,
            primitives,
            &primitives
        );
#if 0
        if (ctx->boundary) {
            bspTree->buildPolygonBoundary();
        }
#endif
        if (options & GLF_OPTION_FLAG(GLF_OCCLUSION_CULL)) {
            bspTree->traverse (
                eye,
                -(float)GLF_EPSILON,
                GLFMath::less,
                GLFViewport::addInImageTree,
                this
            );
            delete imageTree;
            imageTree = 0;
        }
    }
    //
    // process the children viwports
    //
    for (GLFViewport *vp=first; vp; vp=vp->next) {
        vp->process(eye);
    }
}

void GLFViewport::printPrimitives(GLFxyz& eye)
{
    if (!bspTree) {
        process(eye);
    }
    ctx->setOptions(options);
    ctx->writeViewportHeader(viewport,background);
        if (bspTree) {
            bspTree->traverse (
                eye,
                (float)GLF_EPSILON,
                GLFMath::greater,
                GLFContext::printPrimitive,
                ctx
            );
            delete bspTree;
            bspTree = 0;
        }
        primitives.clear();
        //
        // print the children viewports' primitives
        //
        for (GLFViewport *vp=first; vp; vp=vp->next) {
            vp->printPrimitives(eye);
        }
    ctx->writeViewportFooter();
}

void GLFViewport::addInImageTree(GLFPrimitive& prim, void *vp)
{
    prim.culled = !((GLFViewport*)vp)->addInImageTree (
        prim,
        &(((GLFViewport*)vp)->imageTree)
    );
}

void GLFViewport::addPlanesInImageTree(GLFPrimitive& prim, GLFBspTree2D **tree)
{
GLint ret = 0;
size_t i;
size_t offset = 0;
GLFBspTree2D *head = NULL, *cur = NULL;

    if ((*tree == NULL) && (prim.verts.size() > 2)) {
        head = new GLFBspTree2D();
        for (i=0; i<prim.verts.size()-1; i++) {
            if (
                !head->plane.fromPoints (
                    prim.verts[i].xyz,
                    prim.verts[i+1].xyz
                )
            ) {
                if (prim.verts.size()-i > 3) {
                    offset++;
                } else {
                    delete head;
                    return;
                }
            } else {
                break;
            }
        }
        for (i=offset+2; i<prim.verts.size(); i++) {
            ret = prim.verts[i].check(head->plane);
            if (ret!=GLF_POINT_COINCIDENT) {
                break;
            }
        }
        switch (ret) {
            case GLF_POINT_INFRONT :
                    cur = head;
                    for (i=offset+1; i<prim.verts.size()-1; i++) {
                        if (!cur->front) {
                            cur->front = new GLFBspTree2D();
                        }
                        if (
                            cur->front->plane.fromPoints (
                                prim.verts[i].xyz,
                                prim.verts[i+1].xyz
                            )
                        ) {
                            cur = cur->front;
                        }
                    }
                    if (!cur->front) {
                        cur->front = new GLFBspTree2D();
                    }
                    if (
                        !cur->front->plane.fromPoints (
                            prim.verts[i].xyz,
                            prim.verts[offset].xyz
                        )
                    ) {
                        delete cur->front;
                        cur->front = NULL;
                    }
                break;
            case GLF_POINT_BACK :
                for (i=0; i<4; i++) {
                    head->plane[i] = -head->plane[i];
                }
                cur = head;
                for (i=offset+1; i<prim.verts.size()-1; i++) {
                    if (!cur->front) {
                        cur->front = new GLFBspTree2D();
                    }
                    if (
                        cur->front->plane.fromPoints (
                            prim.verts[i+1].xyz,
                            prim.verts[i].xyz
                        )
                    ) {
                        cur = cur->front;
                    }
                }
                if (!cur->front) {
                    cur->front = new GLFBspTree2D();
                }
                if (
                    !cur->front->plane.fromPoints (
                        prim.verts[offset].xyz,
                        prim.verts[i].xyz
                    )
                ) {
                    delete cur->front;
                    cur->front = NULL;
                }
                break;
            default:
                delete head;
                return;
        }
        (*tree) = head;
    }
}

bool GLFViewport::addInImageTree(GLFPrimitive& prim, GLFBspTree2D **tree)
{
GLint ret = 0;
GLFPrimitive *frontprim = NULL;
GLFPrimitive *backprim  = NULL;

    //
    // FIXME: until we consider the actual extent of pixmaps, never cull them.
    // Otherwise the whole pixmap gets culled as soon as the reference point
    // is hidden
    //
    if (prim.type != GLF_PIXMAP) {
        if (!(*tree)) {
            addPlanesInImageTree(prim,tree);
            ret = 1;
        } else {
            switch (prim.check((*tree)->plane)) {
                case GLF_IN_BACK_OF:
                        ret = addInImageTree(prim,&((*tree)->back));
                    break;
                case GLF_IN_FRONT_OF:
                        ret = ((*tree)->front)
                            ? addInImageTree(prim,&((*tree)->front))
                            : 0;
                    break;
                case GLF_SPANNING:
                        if ((backprim = prim.backPiece2D((*tree)->plane))) {
                            ret = addInImageTree(*backprim,&((*tree)->back));
                            delete backprim;
                        }
                        if (
                            (*tree)->front &&
                            (frontprim = prim.frontPiece2D((*tree)->plane))
                        ) {
                            if (addInImageTree(*frontprim,&((*tree)->front))) {
                                ret = 1;
                            }
                            delete frontprim;
                        }
                    break;
                case GLF_COINCIDENT:
                        ret = (prim.verts.size()<3) ? 1 : 0;
                    break;
            }
        }
    } else {
        ret = 1;
    }
    return ret;
}
